#!/usr/bin/env node

// src/cli.ts
import cac from "cac";
import { bundleRequire } from "bundle-require";

// src/util.ts
import escalade from "escalade/sync";
import { mkdirSync } from "fs";
import path from "path";
function getOutputFile(file, format) {
  const id = Math.random().toString(36).substring(2, 15);
  const cacheDir = path.join(nearestNodeModules(file), ".cache/codegentool");
  const outputFile = path.join(
    cacheDir,
    file.replace(
      /\.([mc]?[tj]s|[tj]sx)$/,
      `.bundled_${id}.${format === "esm" ? "mjs" : "cjs"}`
    )
  );
  mkdirSync(path.dirname(outputFile), { recursive: true });
  return outputFile;
}
function nearestNodeModules(file) {
  const dir = escalade(path.dirname(file), (dir2, names) => {
    if (names.includes("node_modules")) {
      return dir2;
    }
  });
  if (dir) {
    return path.join(dir, "node_modules");
  }
  return "node_modules";
}

// src/cli.ts
import { watch } from "chokidar";

// src/config.ts
import escalade2 from "escalade/sync";
import fs from "fs";
function readConfig(cwd) {
  let config;
  let configPath;
  const watchPaths = [];
  escalade2(cwd, (dir, names) => {
    watchPaths.push(dir + "/package.json");
    if (names.includes("package.json")) {
      let pkg;
      try {
        pkg = JSON.parse(fs.readFileSync(dir + "/package.json", "utf8"));
      } catch {
      }
      if (pkg?.codegentool && typeof pkg.codegentool === "object") {
        config = pkg.codegentool;
        configPath = dir + "/package.json";
        return dir;
      }
    }
    if (names.includes(".git")) {
      return dir;
    }
  });
  config = {
    generators: ["generators/**/*.{ts,mts,js,mjs}", "!**/node_modules/**"],
    format: "cjs",
    ...config
  };
  return {
    config,
    configPath,
    watchPaths
  };
}

// src/cli.ts
import { dequal } from "dequal";
import fs2 from "fs";
import path2 from "path";
import resolve from "resolve";
import glob from "fast-glob";
import dedent from "dedent";
import builtinModules from "builtin-modules";
import kleur from "kleur";
var program = cac("codegentool");
program.command("").option("--watch, -w", "Enable file watching and automatic reruns").action((options) => start(process.cwd(), options));
program.parse();
async function start(cwd, options) {
  const { config, configPath, watchPaths } = readConfig(cwd);
  if (configPath) {
    process.chdir(path2.dirname(configPath));
  }
  const watcher = watch([...config.generators, ...watchPaths]);
  const generators = /* @__PURE__ */ new Map();
  const regenerate = debounce((name) => {
    const promise = generators.get(name) ?? Promise.resolve();
    promise.then((close) => close?.()).then(() => {
      generators.set(name, generate(name, config));
    });
  }, 100);
  let initialized = false;
  watcher.on("all", (event, name) => {
    if (name.endsWith("package.json")) {
      if (!initialized)
        return;
      const configResult = readConfig(process.cwd());
      if (!dequal(config, configResult.config) || !dequal(watchPaths, configResult.watchPaths)) {
        watcher.close();
        generators.forEach(async (close) => (await close)());
        start(cwd, options);
      }
    } else if (event == "add" || event == "change") {
      regenerate(name);
    }
  });
  await new Promise((resolve2) => {
    watcher.once("ready", () => {
      initialized = true;
      resolve2();
    });
  });
  if (options.watch) {
    console.log(`\u2318 Watching for changes...`);
  } else {
    watcher.close();
    generators.forEach(async (close) => (await close)());
  }
}
async function generate(generatorPath, config) {
  const watcher = watch([], {
    ignoreInitial: true
  });
  let close = Promise.resolve(() => {
    watcher.close();
  });
  watcher.on(
    "all",
    debounce(() => {
      watcher.close();
      close = generate(generatorPath, config);
    }, 500)
  );
  const resolveModule = (id, importer) => resolve.sync(id, {
    extensions: [".ts", ".js", ".mts", ".cts", ".mjs", ".cjs"],
    basedir: path2.dirname(importer)
  });
  const resolvePlugin = {
    name: "resolve-module",
    setup(build) {
      build.onResolve({ filter: /.*/ }, (args) => {
        if (builtinModules.includes(args.path)) {
          return;
        }
        try {
          return {
            path: resolveModule(args.path, args.importer)
          };
        } catch {
        }
      });
    }
  };
  const requireOptions = {
    tsconfig: config.tsconfig,
    format: config.format,
    esbuildOptions: {
      sourcemap: "inline",
      logLevel: "silent",
      plugins: [resolvePlugin]
    }
  };
  try {
    console.log(
      `\u25B6 Running %s`,
      kleur.cyan(path2.relative(process.cwd(), generatorPath))
    );
    let outputPath;
    const {
      mod: { default: generator },
      dependencies
    } = await bundleRequire({
      ...requireOptions,
      filepath: generatorPath,
      getOutputFile: (file, format) => {
        return outputPath = getOutputFile(file, format);
      }
    });
    if (typeof generator !== "function") {
      throw new Error(`Generator must default export a function`);
    }
    watcher.add(dependencies.filter((file) => file !== generatorPath));
    let filesGenerated = 0;
    const seenErrors = /* @__PURE__ */ new Set();
    const watchPaths = /* @__PURE__ */ new Set();
    const api = {
      scan: (source, options) => {
        const cwd = path2.resolve(options?.cwd || ".");
        if (Array.isArray(source)) {
          source.forEach((s) => watchPaths.add(path2.resolve(cwd, s)));
        } else {
          watchPaths.add(path2.resolve(cwd, source));
        }
        return glob.sync(source, options);
      },
      read: (file, encoding) => {
        watchPaths.add(file);
        return fs2.readFileSync(file, encoding);
      },
      write: (file, data) => {
        if (typeof data === "string") {
          let current;
          try {
            current = fs2.readFileSync(file, "utf8");
          } catch {
          }
          if (data !== current) {
            fs2.mkdirSync(path2.dirname(file), { recursive: true });
            fs2.writeFileSync(file, data);
            filesGenerated++;
            console.log(
              `\u2714\uFE0F Generated %s`,
              kleur.green(path2.relative(process.cwd(), file))
            );
          }
        } else {
          let current;
          try {
            current = fs2.readFileSync(file);
          } catch {
          }
          if (!current?.equals(data)) {
            fs2.mkdirSync(path2.dirname(file), { recursive: true });
            fs2.writeFileSync(file, data);
            filesGenerated++;
            console.log(
              `\u2714\uFE0F Generated %s`,
              kleur.green(path2.relative(process.cwd(), file))
            );
          }
        }
      },
      writeEnv: (file, data) => {
        if (fs2.existsSync(file)) {
          let changed = false;
          const existing = fs2.readFileSync(file, "utf8");
          const existingLines = existing.split("\n");
          const newLines = existingLines.map((line) => {
            const [key, value] = line.split("=");
            if (data[key] !== value) {
              changed = true;
              return `${key}=${data[key].replace(/[\n]/g, "\\n")}`;
            }
            return line;
          });
          if (changed) {
            fs2.writeFileSync(
              file,
              newLines.join("\n") + (existing.endsWith("\n") ? "\n" : "")
            );
            console.log(
              `\u2714\uFE0F Updated env file %s`,
              kleur.green(path2.relative(process.cwd(), file))
            );
          }
        } else {
          fs2.writeFileSync(
            file,
            Object.entries(data).map(([k, v]) => `${k}=${v.replace(/[\n]/g, "\\n")}`).join("\n") + "\n"
          );
          console.log(
            `\u2714\uFE0F Created env file %s`,
            kleur.green(path2.relative(process.cwd(), file))
          );
        }
      },
      dedent,
      loadModule: async (id, basedir) => {
        const isBareSpec = !id.startsWith("./") && !id.startsWith("../");
        try {
          if (!isBareSpec) {
            id = path2.resolve(basedir ?? path2.dirname(generatorPath), id);
          }
          id = resolveModule(id, outputPath);
        } catch (error) {
          console.error(error.message);
          return null;
        }
        try {
          if (isBareSpec) {
            return await import(id);
          } else {
            const { mod, dependencies: dependencies2 } = await bundleRequire({
              ...requireOptions,
              filepath: id,
              getOutputFile
            });
            watcher.add(dependencies2);
            return mod;
          }
        } catch (error) {
          const message = error && (error.stack || error.message) || error;
          if (seenErrors.has(message)) {
            return null;
          }
          seenErrors.add(message);
          console.error(`Error loading module "${id}"`);
          if (error.errors) {
            console.error(error.message);
          } else {
            console.error(error);
          }
          return null;
        }
      }
    };
    await generator(api);
    if (!filesGenerated) {
      console.log(
        `\u2714\uFE0F Nothing updated by %s`,
        kleur.cyan(path2.relative(process.cwd(), generatorPath))
      );
    }
    watcher.add([...watchPaths]);
  } catch (error) {
    console.error(`Error in ${path2.resolve(generatorPath)}`);
    console.error(error);
  }
  return async () => (await close)();
}
function debounce(fn, wait) {
  let timeout;
  return function(...args) {
    clearTimeout(timeout);
    timeout = setTimeout(() => fn.apply(this, args), wait);
  };
}
